/**********************************************************************
 *<
	FILE: surfmod.cpp

	DESCRIPTION:  Varius surface modifiers

	CREATED BY: Rolf Berteig

	HISTORY: 11/07/95

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/
#include "mods.h"
#include "iparamm.h"
#include "edmdata.h"
#include "patch.h"
#include "bmmlib.h"
#include "texutil.h"
#include "mapping.h"
#include "sctex.h"

//#define PATAPP_ID 0x1e3c44,0x273355
#define PATAPP_ID 0x1e3c4455,0x27335544
//#define PATCHMATERIAL_ID 0x1e3c4455,0x27335544

#define PATCHMATERIAL_ID 0xe3c55555, 0x2733665

#define PBLOCK_REF 0

class DataClass
{
public:
	Point3 pos;
	Point2 uv;
	float str;
};

class MatMod : public Modifier {	
	public:
		IParamBlock *pblock;
		static IParamMap *pmapParam;
		Tab<DataClass> data;

		Interface *ip;

		Bitmap *thebm;
		BitmapInfo bi;
		Texmap *tmap;
		float effectstr;
		float blur;

		MatMod();

		// From Animatable
		void DeleteThis() { delete this; }
		void GetClassName(TSTR& s) { s= GetString(IDS_RB_MATMOD); }  
		virtual Class_ID ClassID() { return Class_ID(PATCHMATERIAL_ID);}
		void BeginEditParams( IObjParam  *ip, ULONG flags,Animatable *prev);
		void EndEditParams( IObjParam *ip,ULONG flags,Animatable *next);		
		TCHAR *GetObjectName() { return GetString(IDS_RB_MATERIAL3); }
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;} 

		ChannelMask ChannelsUsed()  {return PART_GEOM|PART_TOPO|PART_SELECT|PART_SUBSEL_TYPE|TEXMAP_CHANNEL;};
		ChannelMask ChannelsChanged() {return GEOM_CHANNEL|TOPO_CHANNEL;}
		
//		Class_ID InputType() {return GEOMOBJECT_CLASS_ID;}

		Class_ID InputType() {return defObjectClassID;}
		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		Interval LocalValidity(TimeValue t);

		int NumRefs() {return 2;}
		RefTargetHandle GetReference(int i);// {return pblock;}
		void SetReference(int i, RefTargetHandle rtarg);// {pblock = (IParamBlock*)rtarg;}

		int NumSubs() {return 2;}
		Animatable* SubAnim(int i);// {return pblock;}
		TSTR SubAnimName(int i);// {return GetString(IDS_RB_PARAMETERS);}

		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message);

		float GetPixel(Point3 pt,float u, float v);
		void FreeBitmap();
		void LoadBitmap();
		void SetImageTime(TimeValue t);

		IOResult Load(ILoad *iload);
		IOResult Save(ISave *isave);

		void EnumAuxFiles(NameEnumCallback& nameEnum, DWORD flags) {
			bi.EnumAuxFiles(nameEnum,flags);
			if (tmap) tmap->EnumAuxFiles(nameEnum,flags);
			}

	};



//--- ClassDescriptor and class vars ---------------------------------

IParamMap *MatMod::pmapParam = NULL;



class MatClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new MatMod; }
	const TCHAR *	ClassName() { return GetString(IDS_RB_MATERIAL3); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(PATCHMATERIAL_ID); }
	const TCHAR* 	Category() { return GetString(IDS_RB_CAT);}
	};

static MatClassDesc matDesc;
extern ClassDesc* GetMatModDesc() { return &matDesc; }


//--- Parameter map/block descriptors -------------------------------

#define PB_MATID 0
#define PB_INCID 1
#define PB_BLUR 2

//
//
// Parameters

static ParamUIDesc descParam[] = {
	
	// Material ID
	ParamUIDesc(
		PB_MATID,
		EDITTYPE_FLOAT,
		IDC_MATID,IDC_MATIDSPIN,
		-100.0f,100.0f,
		0.1f),

	// Inc MatID
	ParamUIDesc(PB_INCID,TYPE_SINGLECHEKBOX,IDC_INC_CHECK),

	
	// Blur
	ParamUIDesc(
		PB_BLUR,
		EDITTYPE_FLOAT,
		IDC_DISP_BLUR,IDC_DISP_BLURSPIN,
		0.0f,10.0f,
		0.01f),


	};

#define PARAMDESC_LENGH 3


static ParamBlockDescID descVer0[] = {
	{ TYPE_FLOAT, NULL, TRUE, 0  },
	{ TYPE_INT, NULL, FALSE, 1 },
	{ TYPE_FLOAT, NULL, TRUE, 2  },
};

#define PBLOCK_LENGTH	3

#define CURRENT_VERSION	0


//--- DispDlgProc -------------------------------------------------------

class MaskDlgProc : public ParamMapUserDlgProc, DADMgr {
	public:
		HWND hWnd;
		MatMod *mod;
//		MatObject *obj;
		MaskDlgProc(MatMod *m) {mod = m;}
//		DispDlgProc(DisplaceObject *o) {mod = NULL;obj = o;}
		BOOL DlgProc(TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);
		void DoBitmapFit(HWND hWnd);
		void SetImageName(HWND hWnd);
		void PickImage(HWND hWnd);
		void RemoveImage(HWND hWnd);
		void SetMapName(HWND hWnd);
		void PickMap(HWND hWnd);
		void RemoveMap(HWND hWnd);
		void DeleteThis() {delete this;}
		void RedrawViews() { 
			Interface *ip = mod->ip;
			ip->RedrawViews(ip->GetTime(),REDRAW_NORMAL);
			}
		// From DADMgr
		SClass_ID GetDragType(HWND hwnd, POINT p) {
			if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKMAP)) 
				return TEXMAP_CLASS_ID;
			if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKIMAGE)) 
				return BITMAPDAD_CLASS_ID;
			else return NULL;
			}		
		BOOL OkToDrop(ReferenceTarget *dropThis, HWND hfrom, HWND hto, POINT p, SClass_ID type, BOOL isNew) {
			if (hfrom==hto) return FALSE;
			if (hto == GetDlgItem(hWnd,IDC_DISP_PICKMAP)) 
				return type==TEXMAP_CLASS_ID;
			if (hto == GetDlgItem(hWnd,IDC_DISP_PICKIMAGE)) 
				return type==BITMAPDAD_CLASS_ID;
			return NULL;
			}
		int SlotOwner() {return OWNER_SCENE;}
		ReferenceTarget *GetInstance(HWND hwnd, POINT p, SClass_ID type) {
			if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKMAP)) {
				return  mod->tmap;
				}
			if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKIMAGE)) {
				DADBitmapCarrier *bmc = GetDADBitmapCarrier();
				BitmapInfo *pbi =  &mod->bi;
				TSTR nm = pbi->Name();
				bmc->SetName(nm);
				return bmc;
				}
			return NULL;
			}
		void Drop(ReferenceTarget *dropThis, HWND hwnd, POINT p, SClass_ID type) {
			if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKMAP)) {
				if (mod) {
					mod->ReplaceReference(2,dropThis);
					mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
					mod->NotifyDependents(FOREVER,0,REFMSG_SUBANIM_STRUCTURE_CHANGED);
				} 	RedrawViews();
				SetMapName(hWnd);
				}
			else if (hwnd == GetDlgItem(hWnd,IDC_DISP_PICKIMAGE)) {
				assert(dropThis->SuperClassID()==BITMAPDAD_CLASS_ID);
				DADBitmapCarrier *bmc = (DADBitmapCarrier*)dropThis;
				if (mod) {
					mod->bi.SetName(bmc->GetName());	
					mod->FreeBitmap();
					mod->LoadBitmap();
					mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
				} 
				RedrawViews();
				SetImageName(hWnd);
				}
			}
		BOOL AutoTooltip() { return TRUE; }
	};


static void GetBMName(BitmapInfo& bi, TSTR &fname) { 
	TSTR fullName;
	if (bi.Name()[0]==0)
		fullName = bi.Device();
	else 
		fullName =  bi.Name();
	SplitPathFile(fullName,NULL,&fname);
	}

void MaskDlgProc::SetImageName(HWND hWnd)
	{ 	
 	if ((mod&&mod->thebm) ) {
		TSTR fname;
		if (mod) 
			GetBMName(mod->bi, fname); 
		SetDlgItemText(hWnd, IDC_DISP_PICKIMAGE,fname);
	} else {
		SetDlgItemText(hWnd, IDC_DISP_PICKIMAGE, GetString(IDS_RB_NONE));
		}
	InvalidateRect(GetDlgItem(hWnd,IDC_DISP_PICKIMAGE),NULL,TRUE);
	}

void MaskDlgProc::PickImage(HWND hWnd)
	{	
	BitmapInfo *bi = &mod->bi;
	if(TheManager->SelectFileInputEx(
		bi, hWnd, GetString(IDS_RB_SELECTDISPIMAGE)))
		{
		if (mod) {
			mod->FreeBitmap();
			mod->LoadBitmap();
			mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
		} 
		SetImageName(hWnd);
		}
	}

void MaskDlgProc::RemoveImage(HWND hWnd)
	{
	if (mod) {
		mod->bi.SetName(_T(""));
		mod->bi.SetDevice(_T(""));
		mod->FreeBitmap();
		mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
	} 
	SetImageName(hWnd);
	}

void MaskDlgProc::SetMapName(HWND hWnd)
	{
	if ((mod&&mod->tmap) ) {
		TSTR fname;
		if (mod)  fname = mod->tmap->GetFullName();
	
		SetDlgItemText(hWnd, IDC_DISP_PICKMAP,fname);
		InvalidateRect(GetDlgItem(hWnd,IDC_DISP_PICKMAP),NULL,TRUE);
	} else {
		SetDlgItemText(hWnd, IDC_DISP_PICKMAP, GetString(IDS_RB_NONE));
		InvalidateRect(GetDlgItem(hWnd,IDC_DISP_PICKMAP),NULL,TRUE);
		}
	}

void MaskDlgProc::PickMap(HWND hWnd)
	{
	BOOL newMat=FALSE, cancel=FALSE;
	Interface *ip = mod->ip;
	MtlBase *mtl = ip->DoMaterialBrowseDlg(
		hWnd,
		BROWSE_MAPSONLY|BROWSE_INCNONE|BROWSE_INSTANCEONLY,
		newMat,cancel);
	if (cancel) {
		if (newMat) mtl->DeleteThis();
		return;
		}
	if (mod) {
		mod->ReplaceReference(1,mtl);
		mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
		mod->NotifyDependents(FOREVER,0,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	} 
	SetMapName(hWnd);
	}

void MaskDlgProc::RemoveMap(HWND hWnd)
	{
	if (mod) {
		mod->ReplaceReference(1,NULL);
		mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
		mod->NotifyDependents(FOREVER,0,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	}
	SetMapName(hWnd);
	}

void MaskDlgProc::DoBitmapFit(HWND hWnd)
	{
	}

BOOL MaskDlgProc::DlgProc(
		TimeValue t,IParamMap *map,
		HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
	{
	switch (msg) {
		case WM_INITDIALOG: {
			this->hWnd = hWnd;
			SetImageName(hWnd);
			SetMapName(hWnd);
			ICustButton *but = GetICustButton(GetDlgItem(hWnd,IDC_DISP_PICKMAP));
			but->SetDADMgr(this);
			ReleaseICustButton(but);
			but = GetICustButton(GetDlgItem(hWnd,IDC_DISP_PICKIMAGE));
			but->SetDADMgr(this);
			ReleaseICustButton(but);
			break;
			}
		
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				
				case IDC_DISP_PICKIMAGE:
					PickImage(hWnd);
					break;
				
				case IDC_DISP_REMOVEIMAGE:
					RemoveImage(hWnd);
					break;

				case IDC_DISP_PICKMAP:
					PickMap(hWnd);
					break;
				
				case IDC_DISP_REMOVEMAP:
					RemoveMap(hWnd);
					break;

				}
			break;
		}
	
	return FALSE;
	}



//--- MatMod methods -------------------------------


MatMod::MatMod()
	{	
	MakeRefByID(FOREVER, 0, 
		CreateParameterBlock(descVer0, PBLOCK_LENGTH, CURRENT_VERSION));	
	pblock->SetValue(PB_MATID,0,1.0f);
	pblock->SetValue(PB_BLUR,0,0.0f);
	pblock->SetValue(PB_INCID,0,1);
	tmap = NULL;
	thebm = NULL;

	}

void MatMod::BeginEditParams(IObjParam *ip, ULONG flags,Animatable *prev)
	{
	this->ip = ip;
	pmapParam = CreateCPParamMap(
		descParam,PARAMDESC_LENGH,
		pblock,
		ip,
		hInstance,
		MAKEINTRESOURCE(IDD_MATERIALPARAM),
		GetString(IDS_RB_PARAMETERS),
		0);	
	pmapParam->SetUserDlgProc(new MaskDlgProc(this));	
	}
		
void MatMod::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
	{	
	DestroyCPParamMap(pmapParam);
	}

Interval MatMod::LocalValidity(TimeValue t)
	{
	int i;
	float f;
	Interval valid = FOREVER;
	pblock->GetValue(PB_MATID,t,f,valid);	
	pblock->GetValue(PB_BLUR,t,f,valid);	
	if (thebm) {
		if (bi.FirstFrame()!=bi.LastFrame()) {
			valid.SetInstant(t);
			return valid;
			}
		}
	if (tmap) {
		valid &= tmap->Validity(t);
		tmap->LoadMapFiles(t);

		}

//   valid.Set(t,t);
	return valid;
	}

RefTargetHandle MatMod::Clone(RemapDir& remap) 
	{
	MatMod* newmod = new MatMod();	
	newmod->ReplaceReference(0,pblock->Clone(remap));	
//	newmod->ReplaceReference(PBLOCK_REF,pblock->Clone(remap));	
	newmod->bi = bi;
	if (bi.Name()[0]!=0||bi.Device()[0]!=0) {
		newmod->LoadBitmap();
		}

	return newmod;
	}


static void SetTMapTime(TimeValue t,Texmap *tmap)
	{
	if (!tmap) return;
	tmap->LoadMapFiles(t);
	tmap->Update(t,FOREVER);
	for (int i=0; i<tmap->NumSubTexmaps(); i++) {
		SetTMapTime(t,tmap->GetSubTexmap(i));
		}
	}


void MatMod::SetImageTime(TimeValue t)
	{
	if (thebm) {
		int length = bi.LastFrame()-bi.FirstFrame()+1;
		int frame = (t/GetTicksPerFrame())%length;
		if (frame<0) frame += length;
		frame += bi.FirstFrame();
		bi.SetCurrentFrame(frame);
		thebm->GoTo(&bi);
		}
	if (tmap) {
		SetTMapTime(t,tmap);
		}
	}

void MatMod::LoadBitmap() 
	{
	if (bi.Name()[0]==0&&bi.Device()[0]==0)
		return;
	thebm = TheManager->Load(&bi);
	}

void MatMod::FreeBitmap()
	{
	if (thebm) {
		thebm->DeleteThis();
		thebm = NULL;
		}
	}


float MatMod::GetPixel(Point3 pt,float u, float v)
	{
	if (!thebm && !tmap) return 1.0f;
	float f = 0.0f;
//	float blur  = 0.0f;
	if (thebm) {
		BMM_Color_64 c;
		if (blur==0.0f) {
			int w = thebm->Width()-1;
			int h = thebm->Height()-1;
			int x = (int)(frac(u)*(float)w);
			int y = h - (int)(frac(v)*(float)h);
			thebm->GetLinearPixels(x,y,1,&c);
		} else {		
			float wb = (blur*10.0f+1.0f)/float(thebm->Width());
			float hb = (blur*10.0f+1.0f)/float(thebm->Height());
			thebm->GetFiltered(frac(u),1.0f-frac(v),wb,hb,&c);
			}
		f += float(c.r)/float(0xFFFF) * 0.299f;
		f += float(c.g)/float(0xFFFF) * 0.587f;
		f += float(c.b)/float(0xFFFF) * 0.114f;		
		} 
	if (tmap) {
		SCTex shadeContext;
		shadeContext.scale = 1.0f;
		shadeContext.duvw  = Point3(blur/10.0f+0.001f,blur/10.0f+0.001f,0.0f);
		shadeContext.scrPos.x = int(u);
		shadeContext.scrPos.y = int(v);
		shadeContext.uvw.x = u;
		shadeContext.uvw.y = v;
		shadeContext.uvw.z = 0.0f;
		shadeContext.pt = pt;
		shadeContext.dpt = Point3(blur/10.0f+0.001f,blur/10.0f+0.001f,blur/10.0f+0.001f);
		AColor c;
		c  = tmap->EvalColor(shadeContext);
		f += c.r * 0.299f;
		f += c.g * 0.587f;
		f += c.b * 0.114f;		
		}
	return f;
	}


class MatDeformer : public Deformer {
	public:		

		MatMod *mod;
		MatDeformer(MatMod *m) {mod = m;}
		Point3 Map(int i, Point3 p);
	};


Point3 MatDeformer::Map(int i, Point3 p)
	{
	if (i < mod->data.Count())
		p = ((p - mod->data[i].pos) * mod->data[i].str * mod->effectstr) + mod->data[i].pos;
	return p;
	}




void MatMod::ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node)
	{
	Interval valid = FOREVER;
	pblock->GetValue(PB_MATID,t,effectstr,valid);	
	pblock->GetValue(PB_BLUR,t,blur,valid);	
//	id--;
//	if (id<0) id = 0;
//	if (id>0xffff) id = 0xffff;
	



		
//	MtlID matID = id;
//	MtlID *matIDList,*tmat;




	int len = os->obj->NumPoints();

	Animatable *anim = os->obj;
	SClass_ID scid = anim->SuperClassID();

DebugPrint("Mask Top %d\n",t);
/*
	// Grab the app data chunk from the anim
	AppDataChunk *ad = 
		anim->GetAppDataChunk(
			Class_ID(APPMATERIAL_ID), 
			GEOMOBJECT_CLASS_ID, 
			0);



*/

	// Grab the app data chunk from the anim
	AppDataChunk *ad = 
		anim->GetAppDataChunk(
			Class_ID(PATAPP_ID), 
			scid, 
			0);

	Point3 *PointList = NULL;
	if (ad) {
		// Fill the edit field with the text.
			PointList = (Point3*)ad->data;
DebugPrint("         Mask Top Data found  %d\n",t);
			}

/*
	if (ad) {
		// Fill the edit field with the text.
		tdata = (DataClass*)ad->data;
		data = (DataClass*)malloc(len*sizeof(DataClass));
		for (int i=0; i<len; i++) {
			data[i]=tdata[i];
			}

DebugPrint("Old data present\n");
	} else {
		// Not found!
		// data create
DebugPrint("Creating New data present len %d\n",len);
		data = (DataClass*)malloc(len*sizeof(DataClass));

//		MessageBox(hPanel,_T("No app data found"),_T("App Data Tester"),MB_OK);
		}
*/
	if (PointList != NULL)
		{

		TriObject *tobj = (TriObject*)os->obj;
		Mesh &mesh = tobj->mesh;

		data.ZeroCount();
		data.SetCount(len);
//get point
//ge	t uv
 		int tvcount = mesh.getNumTVerts();
		DebugPrint("         Tverts Count %d\n",tvcount);

		if (tvcount > 0)
			{
			for (int i=0; i<len; i++) 
				{	
				data[i].pos = PointList[i];
				data[i].uv = Point2(0.0f,0.0f); 
				data[i].str = 0.0f; 
				}

			for (i=0; i<mesh.getNumFaces(); i++) 
				{

				for (int j=0; j<3; j++) {
					    int tvf = mesh.tvFace[i].t[j];
						if (tvf < tvcount)
							{
							Point3 pt = mesh.tVerts[mesh.tvFace[i].t[j]];
							data[mesh.faces[i].v[j]].uv = Point2(pt.x,pt.y);
							data[mesh.faces[i].v[j]].str = GetPixel(data[mesh.faces[i].v[j]].pos,pt.x,pt.y);
							}
						}
				}
			}
			
		}

	MatDeformer deformer(this);
	os->obj->Deform(&deformer, TRUE);
	os->obj->UpdateValidity(GEOM_CHAN_NUM, LocalValidity(t));

/*
// Remove the chunk if it exists
	anim->RemoveAppDataChunk(
			Class_ID(PATCHMATERIAL_ID), 
			GEOMOBJECT_CLASS_ID, 
			0);

//put appdata back
	// Add a chunk
	anim->AddAppDataChunk(
			Class_ID(PATCHMATERIAL_ID), 
			GEOMOBJECT_CLASS_ID, 
			0,
			(DWORD)len*sizeof(DataClass), data);
*/
	}


RefTargetHandle MatMod::GetReference(int i)
	{
	switch (i) {
		case PBLOCK_REF:	return pblock;
		case 1:             return tmap;
		default: 			return NULL;
		}
	}

void MatMod::SetReference(int i, RefTargetHandle rtarg)
	{
	switch (i) {
		case PBLOCK_REF:	pblock = (IParamBlock*)rtarg; break;
		case 1:             
			tmap = (Texmap*)rtarg; 
			if (( pmapParam) && (tmap != NULL)){
				((MaskDlgProc*)pmapParam->GetUserDlgProc())->SetMapName(
					pmapParam->GetHWnd());
				}
			break;
		}
	}

Animatable* MatMod::SubAnim(int i)
	{
	switch (i) {
		case PBLOCK_REF:	return pblock;
		case 1:             return tmap;
		default: 			return NULL;   
		}
	}

TSTR MatMod::SubAnimName(int i)
	{
	switch (i) {
		case PBLOCK_REF:	return GetString(IDS_RB_PARAMETERS);
		case 1:             return _T("Mask Map");
		default: 			return TSTR(_T(""));
		}
	}



RefResult MatMod::NotifyRefChanged(
		Interval changeInt, 
		RefTargetHandle hTarget, 
   		PartID& partID, 
   		RefMessage message) 
   	{
	switch (message) {
		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {				
				case 0:
				default: gpd->dim = defaultDim; break;
				}			
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			switch (gpn->index) {				
				case PB_MATID:	gpn->name = TSTR(_T("Effect Strength")); break;
				case PB_BLUR:	gpn->name = TSTR(_T("Blur")); break;
				default:		gpn->name = TSTR(_T("")); break;
				}
			return REF_STOP; 
			}
		}
	return REF_SUCCEED;
	}



#define MAPNAME_CHUNK	0x0100
#define BMIO_CHUNK	0x0200
#define NEWMAP_CHUNKID	0x0300

IOResult MatMod::Load(ILoad *iload)
	{	
	IOResult res = IO_OK;
	Modifier::Load(iload);

//	iload->RegisterPostLoadCallback(
//		new FixDimsPLCB(this));

//	iload->RegisterPostLoadCallback(
//		new ParamBlockPLCB(versions,NUM_OLDVERSIONS,&curVersion,this,PBLOCK_REF));

	ClearAFlag(A_PLUGIN1);

	while (IO_OK==(res=iload->OpenChunk())) {
		switch (iload->CurChunkID()) {
			case MAPNAME_CHUNK: {
				TCHAR *buf;
				res=iload->ReadWStringChunk(&buf);
				bi.SetName(buf);
				FreeBitmap();
				LoadBitmap();
				break;
				}
			case BMIO_CHUNK: 
				res = bi.Load(iload);
				FreeBitmap();
				LoadBitmap();
				break;

			}
		iload->CloseChunk();
		if (res!=IO_OK)  return res;
		}	
	
	return IO_OK;
	}

IOResult MatMod::Save(ISave *isave)
	{
	Modifier::Save(isave);
	isave->BeginChunk(BMIO_CHUNK);
	bi.Save(isave);
	isave->EndChunk();
	return IO_OK;
	}

