/**********************************************************************
 *<
	FILE: surfmod.cpp

	DESCRIPTION:  Varius surface modifiers

	CREATED BY: Rolf Berteig

	HISTORY: 11/07/95

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/
#include "mods.h"
#include "iparamm.h"
#include "edmdata.h"
#include "patch.h"


#define PATCHMATERIAL_ID 0x1e3c4455,0x27335544
//#define PATAPP_ID 0x1e3c44,0x273355


class MatMod : public Modifier {	
	public:
		IParamBlock *pblock;
		static IParamMap *pmapParam;



		MatMod();

		// From Animatable
		void DeleteThis() { delete this; }
		void GetClassName(TSTR& s) { s= GetString(IDS_RB_MATMOD); }  
		virtual Class_ID ClassID() { return Class_ID(PATCHMATERIAL_ID);}
		void BeginEditParams( IObjParam  *ip, ULONG flags,Animatable *prev);
		void EndEditParams( IObjParam *ip,ULONG flags,Animatable *next);		
		TCHAR *GetObjectName() { return GetString(IDS_RB_MATERIAL3); }
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;} 

		ChannelMask ChannelsUsed()  {return OBJ_CHANNELS;}
		ChannelMask ChannelsChanged() {return GEOM_CHANNEL|TOPO_CHANNEL;}
		
//		Class_ID InputType() {return GEOMOBJECT_CLASS_ID;}

		Class_ID InputType() {return defObjectClassID;}
		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		Interval LocalValidity(TimeValue t);

		int NumRefs() {return 1;}
		RefTargetHandle GetReference(int i) {return pblock;}
		void SetReference(int i, RefTargetHandle rtarg) {pblock = (IParamBlock*)rtarg;}

		int NumSubs() {return 1;}
		Animatable* SubAnim(int i) {return pblock;}
		TSTR SubAnimName(int i) {return GetString(IDS_RB_PARAMETERS);}

		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message);
	};



//--- ClassDescriptor and class vars ---------------------------------

IParamMap *MatMod::pmapParam = NULL;



class MatClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new MatMod; }
	const TCHAR *	ClassName() { return GetString(IDS_RB_MATERIAL3); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(PATCHMATERIAL_ID); }
	const TCHAR* 	Category() { return GetString(IDS_RB_CAT);}
	};

static MatClassDesc matDesc;
extern ClassDesc* GetMatModDesc() { return &matDesc; }


//--- Parameter map/block descriptors -------------------------------

#define PB_MATID 0
#define PB_INCID 1
#define PB_ONEMAP 2

//
//
// Parameters

static ParamUIDesc descParam[] = {
	
	// Material ID
	ParamUIDesc(
		PB_MATID,
		EDITTYPE_INT,
		IDC_MATID,IDC_MATIDSPIN,
		1.0f,(float)0xffff,
		0.1f),

	// Inc MatID
	ParamUIDesc(PB_INCID,TYPE_SINGLECHEKBOX,IDC_INC_CHECK)
		

	};
#define PARAMDESC_LENGH 2


static ParamBlockDescID descVer0[] = {
	{ TYPE_INT, NULL, TRUE, 0  },
	{ TYPE_INT, NULL, FALSE, 1 },
};
#define PBLOCK_LENGTH	2

#define CURRENT_VERSION	0


//--- MatMod methods -------------------------------


MatMod::MatMod()
	{	
	MakeRefByID(FOREVER, 0, 
		CreateParameterBlock(descVer0, PBLOCK_LENGTH, CURRENT_VERSION));	
	pblock->SetValue(PB_MATID,0,1);
	pblock->SetValue(PB_INCID,0,1);
	}

void MatMod::BeginEditParams(IObjParam *ip, ULONG flags,Animatable *prev)
	{
	pmapParam = CreateCPParamMap(
		descParam,PARAMDESC_LENGH,
		pblock,
		ip,
		hInstance,
		MAKEINTRESOURCE(IDD_MATERIALPARAM),
		GetString(IDS_RB_PARAMETERS),
		0);		
	}
		
void MatMod::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
	{	
	DestroyCPParamMap(pmapParam);
	}

Interval MatMod::LocalValidity(TimeValue t)
	{
	int i;
	Interval valid = FOREVER;
	pblock->GetValue(PB_MATID,t,i,valid);	
 	valid.Set(t,t);
	return valid;
	}

RefTargetHandle MatMod::Clone(RemapDir& remap) 
	{
	MatMod* newmod = new MatMod();	
	newmod->ReplaceReference(0,pblock->Clone(remap));	
	return newmod;
	}

void MatMod::ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node)
	{
	Interval valid = FOREVER;
	int len = os->obj->NumPoints();



//check app data exists
	Animatable *anim = os->obj;
	SClass_ID scid = anim->SuperClassID();
	// Grab the app data chunk from the anim
	AppDataChunk *ad = 
		anim->GetAppDataChunk(
			Class_ID(PATCHMATERIAL_ID), 
			scid, 
			0);

	Point3 *PointList = NULL;
	Point3 *TPointList = NULL;
DebugPrint("Mask Bottom %d\n",t);
	if (ad) {
		// Fill the edit field with the text.
// Remove the chunk if it exists
		anim->RemoveAppDataChunk(
			Class_ID(PATCHMATERIAL_ID), 
			scid, 
			0);

//		TPointList = (Point3*)ad->data;
		PointList= (Point3*)malloc(len*sizeof(Point3));

DebugPrint("Bottom Old data present\n");
	} else {
		// Not found!
		// data create
DebugPrint("Bottom  Creating New data present len %d\n",len);
		PointList = (Point3*)malloc(len*sizeof(Point3));

//		MessageBox(hPanel,_T("No app data found"),_T("App Data Tester"),MB_OK);
		}


//if exists get it
//else create it
DebugPrint("Dumping New data present len %d\n",len);
	for (int i=0; i<len; i++) {
		PointList[i]=os->obj->GetPoint(i);
		}

//put appdata back
// Add a chunk
	anim->AddAppDataChunk(
			Class_ID(PATCHMATERIAL_ID), 
			scid, 
			0,
			(DWORD)len*sizeof(Point3), PointList);

	os->obj->UpdateValidity(GEOM_CHAN_NUM, LocalValidity(t));


							   

	}

RefResult MatMod::NotifyRefChanged(
		Interval changeInt, 
		RefTargetHandle hTarget, 
   		PartID& partID, 
   		RefMessage message) 
   	{
	switch (message) {
		case REFMSG_GET_PARAM_DIM: {
			GetParamDim *gpd = (GetParamDim*)partID;
			switch (gpd->index) {				
				case 0:
				default: gpd->dim = defaultDim; break;
				}			
			return REF_STOP; 
			}

		case REFMSG_GET_PARAM_NAME: {
			GetParamName *gpn = (GetParamName*)partID;
			switch (gpn->index) {				
				case PB_MATID:	gpn->name = GetString(IDS_RB_MATERIALID); break;
				default:		gpn->name = TSTR(_T("")); break;
				}
			return REF_STOP; 
			}
		}
	return REF_SUCCEED;
	}

